# Filename: pickle.py
# pylint: disable=locally-disabled
"""
Pump for the pickle data format.

"""
from __future__ import absolute_import, print_function, division

import pickle
from collections import namedtuple

from km3pipe.core import Pump
from km3pipe.logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)    # pylint: disable=C0103


class PicklePump(Pump):
    """A pump for Python pickles."""

    def configure(self):

        self.filename = self.get('filename')
        self.events = pickle.load(open(self.filename, "rb"))

    def get_blob(self, index):
        hits = self.events[0]
        t0 = min([hit[3] for hit in hits])
        raw_hits = []
        for hit in hits:
            raw_hit = RawHit(hit[0], hit[1], hit[2], hit[3] - t0)
            print(raw_hit)
            raw_hits.append(raw_hit)
        blob = {'EvtRawHits': raw_hits}
        return blob


RawHit = namedtuple('RawHit', 'id pmt_id tot time')



# Filename: no_test_daq.py
# pylint: disable=C0111,R0904,C0103
"""
Tests for KM3NeT binary formats readout.

"""
from os.path import dirname, join

from km3pipe.testing import TestCase
from km3pipe.io.daq import (DAQPump, DAQPreamble, DAQHeader, DAQSummaryslice)

TEST_DATA_DIR = join(dirname(__file__), "../../kp-data/test_data")
IO_SUM_FILE = join(TEST_DATA_DIR, "IO_SUM.dat")
IO_EVT_FILE = join(TEST_DATA_DIR, "IO_EVT.dat")


class TestDAQPump(TestCase):
    def test_init(self):
        DAQPump()

    def test_init_with_filename(self):
        DAQPump(IO_SUM_FILE)

    def test_frame_positions_in_io_sum(self):
        p = DAQPump(IO_SUM_FILE)
        assert 81 == len(p.frame_positions)
        self.assertListEqual([0, 656, 1312], p.frame_positions[:3])
        self.assertListEqual([50973, 51629, 52285], p.frame_positions[-3:])

    def test_frame_positions_in_io_evt(self):
        p = DAQPump(IO_EVT_FILE)
        assert 38 == len(p.frame_positions)
        self.assertListEqual([0, 570, 986], p.frame_positions[:3])
        self.assertListEqual([13694, 14016, 14360], p.frame_positions[-3:])

    def test_blob_in_io_sum(self):
        p = DAQPump(IO_SUM_FILE)
        blob = p.next_blob()
        assert 'DAQSummaryslice' in blob.keys()
        assert 'DAQPreamble' in blob.keys()
        assert 'DAQHeader' in blob.keys()
        assert 16 == blob['DAQSummaryslice'].n_summary_frames

    def test_blob_in_io_evt(self):
        p = DAQPump(IO_EVT_FILE)
        blob = p.next_blob()
        assert 'DAQEvent' in blob.keys()
        assert 'DAQPreamble' in blob.keys()
        assert 'DAQHeader' in blob.keys()
        event = blob['DAQEvent']
        assert 13 == event.n_triggered_hits
        assert 28 == event.n_snapshot_hits

    def test_blob_iteration(self):
        p = DAQPump(IO_EVT_FILE)
        for blob in p:
            pass

    def test_get_item(self):
        p = DAQPump(IO_EVT_FILE)
        blob = p[4]
        event = blob['DAQEvent']
        assert 6 == event.n_triggered_hits
        assert 17 == event.n_snapshot_hits



#!/usr/bin/env python
"""
======================
ToHDF5 Pipeline example
======================

"""

# Authors: Moritz Lotze <mlotze@km3net.de>
# License: BSD-3
# Date: 2018-07-12
# Status: Under construction...

from __future__ import absolute_import, print_function, division

import tables as tb

from km3pipe import Pipeline
from km3pipe.io import EvtPump, HDF5Sink

from km3modules.common import StatusBar

#####################################################
# Preparation
# -----------
# Let's define the inputs / outputs first -- those would be coming from a CLI
# parser in practice.

N_EVENTS = 200000
IN_FNAME = 'data/numu_cc.evt'
OUT_FNAME = 'data/numu_cc.h5'

#####################################################
# Also, in this case we don't really want to dump the data onto disk, so we
# create an in-memory-ony file, and pass it as the ``h5file`` arg to the
# hdf5sink. to actually write out a file, just specify an outfile name
# (commented out here).

OUTFILE = tb.open_file(
    # create the file in memory only
    OUT_FNAME,
    'w',
    driver="H5FD_CORE",
    driver_core_backing_store=0,
)

#####################################################
# Setting up the pipeline
# -----------------------

pipe = Pipeline(timeit=True)
pipe.attach(EvtPump, filename=IN_FNAME)
pipe.attach(StatusBar, every=25)
pipe.attach(
    HDF5Sink,
    # filename=OUT_FNAME,
    h5file=OUTFILE,
)

#####################################################
# Draining the pipeline
# ---------------------

pipe.drain(N_EVENTS)



# Filename: test_srv.py
# pylint: disable=locally-disabled,C0111,R0904,C0103
from km3pipe.testing import TestCase, patch
from km3pipe.dataclasses import Table
from km3pipe.srv import srv_event

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestSrvEvent(TestCase):
    @patch('km3pipe.srv.srv_data')
    def test_call(self, srv_data_mock):
        hits = Table({
            'pos_x': [1, 2],
            'pos_y': [3, 4],
            'pos_z': [5, 6],
            'time': [100, 200],
            'tot': [11, 22]
        })
        srv_event('token', hits, 'rba_url')
        srv_data_mock.assert_called_once()



# Filename: test_io.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
import tempfile

import tables

from km3pipe.tools import istype
from km3pipe.testing import TestCase, patch, Mock
from km3pipe.io import GenericPump, read_calibration

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestGenericPump(TestCase):
    def test_init_raies_filenotfounderror_for_nonexistinent_file(self):
        with self.assertRaises(ValueError):
            GenericPump("unsupport-file")

    def test_filenames_should_be_valid_iterable(self):
        with self.assertRaises(TypeError):
            GenericPump(None)

    def test_init_raises_ioerror_for_mixed_filetypes(self):
        with self.assertRaises(IOError):
            GenericPump(['a.a', 'b.b'])

    def test_init_evt_with_one_file(self):
        fobj = tempfile.NamedTemporaryFile(delete=True, suffix='.evt')
        fname = str(fobj.name)
        p = GenericPump(fname)
        fobj.close()
        assert istype(p, 'EvtPump')
        assert fname == p.filename

    def test_init_h5_with_one_file(self):
        fobj = tempfile.NamedTemporaryFile(delete=True, suffix='.h5')
        fname = str(fobj.name)
        # We expect just an HDF5ExtError if "everything" goes well
        with self.assertRaises(tables.HDF5ExtError):
            GenericPump(fname)
        fobj.close()

    def test_init_h5_with_multiple_files(self):
        fobj1 = tempfile.NamedTemporaryFile(delete=True, suffix='.h5')
        fobj2 = tempfile.NamedTemporaryFile(delete=True, suffix='.h5')
        fname1 = str(fobj1.name)
        fname2 = str(fobj2.name)
        # We expect just an HDF5ExtError if "everything" goes well
        with self.assertRaises(tables.HDF5ExtError):
            GenericPump([fname1, fname2])
        fobj1.close()
        fobj2.close()

    def test_init_h5_with_multiple_files_where_one_nonexistent(self):
        fobj1 = tempfile.NamedTemporaryFile(delete=True, suffix='.h5')
        fobj2 = tempfile.NamedTemporaryFile(delete=True, suffix='.h5')
        fname1 = str(fobj1.name)
        fname2 = str(fobj2.name)
        fname3 = "nonexistent-file.h5"
        # We expect just an HDF5ExtError if "everything" goes well
        with self.assertRaises(tables.HDF5ExtError):
            GenericPump([fname1, fname2, fname3])
        fobj1.close()
        fobj2.close()

    def test_init_h5_with_nonexistent_files(self):
        fnames = ["nonexistent-file{}.h5".format(i) for i in range(3)]
        with self.assertRaises(SystemExit):
            GenericPump(fnames)


class TestReadCalibration(TestCase):
    def test_call(self):
        read_calibration()

    @patch('km3pipe.calib.Calibration')
    def test_call_with_detx(self, mock_calibration):
        read_calibration(detx='a')
        mock_calibration.assert_called_with(filename='a')

    @patch('km3pipe.calib.Calibration')
    def test_call_with_det_id(self, mock_calibration):
        det_id = 1
        read_calibration(det_id=det_id)
        mock_calibration.assert_called_with(det_id=det_id)

    @patch('km3pipe.calib.Calibration')
    def test_call_with_negative_det_id(self, mock_calibration):
        det_id = -1
        read_calibration(det_id=det_id)
        mock_calibration.assert_not_called()



# Filename: test_hardware.py
# pylint: disable=C0111,C0103,R0904
"""
Detector description (detx format v5)

global_det_id ndoms
dom_id line_id floor_id npmts
 pmt_id_global x y z dx dy dz t0
 pmt_id_global x y z dx dy dz t0
 ...
 pmt_id_global x y z dx dy dz t0
dom_id line_id floor_id npmts
 ...

"""
from __future__ import unicode_literals

from copy import deepcopy
from os.path import join, dirname
from io import StringIO

import numpy as np

from km3pipe.testing import TestCase
from km3pipe.hardware import Detector, PMT
from km3pipe.math import qrot_yaw

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

EXAMPLE_DETX = StringIO(
    "\n".join((
        "1 3",
        "1 1 1 3",
        " 1 1.1 1.2 1.3 -1.1  0.2  0.3 10",
        " 2 1.4 1.5 1.6  0.1 -1.2  0.3 20",
        " 3 1.7 1.8 1.9  0.1  0.2 -1.3 30",
        "2 1 2 3",
        " 4 2.1 2.2 2.3 -1.1  0.2  0.3 40",
        " 5 2.4 2.5 2.6  0.1 -1.2  0.3 50",
        " 6 2.7 2.8 2.9  0.1  0.2 -1.3 60",
        "3 1 3 3",
        " 7 3.1 3.2 3.3 -1.1  0.2  0.3 70",
        " 8 3.4 3.5 3.6  0.1 -1.2  0.3 80",
        " 9 3.7 3.8 3.9  0.1  0.2 -1.3 90",
    ))
)

EXAMPLE_DETX_MIXED_IDS = StringIO(
    "\n".join((
        "1 3",
        "8 1 1 3",
        " 83 1.1 1.2 1.3 -1.1  0.2  0.3 10",
        " 81 1.4 1.5 1.6  0.1 -1.2  0.3 20",
        " 82 1.7 1.8 1.9  0.1  0.2 -1.3 30",
        "7 1 2 3",
        " 71 2.1 2.2 2.3 -1.1  0.2  0.3 40",
        " 73 2.4 2.5 2.6  0.1 -1.2  0.3 50",
        " 72 2.7 2.8 2.9  0.1  0.2 -1.3 60",
        "6 1 3 3",
        " 62 3.1 3.2 3.3 -1.1  0.2  0.3 70",
        " 63 3.4 3.5 3.6  0.1 -1.2  0.3 80",
        " 61 3.7 3.8 3.9  0.1  0.2 -1.3 90",
    ))
)

EXAMPLE_DETX_RADIAL = StringIO(
    "\n".join((
        "1 3",
        "1 1 1 4",
        " 1 1 0 0 1 0 0 10",
        " 2 0 1 0 0 1 0 20",
        " 3 -1 0 0 -1 0 0 30",
        " 4 0 -1 0 0 -1 0 40",
        "2 1 2 2",
        " 5 0 0 1 0 0 1 50",
        " 6 0 0 -1 0 0 -1 60",
        "3 1 3 2",
        " 7 1 2 3 1 2 3 70",
        " 8 -3 -2 -1 -3 -2 -1 80",
        "4 2 1 2",
        " 9 0 0 1 0 0 1 90",
        " 10 0 0 -1 0 0 -1 100",
    ))
)

TEST_DATA_DIR = join(dirname(__file__), "../kp-data/test_data")


class TestDetector(TestCase):
    def setUp(self):
        self.det = Detector()
        self.det._det_file = EXAMPLE_DETX

    def test_parse_header_extracts_correct_det_id(self):
        self.det._parse_header()
        self.assertEqual(1, self.det.det_id)

    def test_parse_header_extracts_correct_n_doms(self):
        self.det._parse_header()
        self.assertEqual(3, self.det.n_doms)

    def test_parse_doms_maps_each_dom_correctly(self):
        self.det._parse_doms()
        expected = {1: (1, 1, 3), 2: (1, 2, 3), 3: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_dom_ids(self):
        self.det._parse_doms()
        self.assertEqual((1, 2, 3), tuple(self.det.dom_ids))

    def test_parse_reset_cache(self):
        self.det._parse_doms()
        assert not self.det._dom_positions
        assert not self.det._pmt_angles
        assert not self.det._xy_positions
        self.det.dom_positions
        self.det.pmt_angles
        self.det.xy_positions
        assert self.det._dom_positions
        assert len(self.det._pmt_angles) == 3
        assert len(self.det._xy_positions) == 1
        self.det.reset_caches()
        assert not self.det._dom_positions
        assert not self.det._pmt_angles
        assert not self.det._xy_positions

    def test_parse_doms_maps_each_dom_correctly_for_mixed_pmt_ids(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        expected = {8: (1, 1, 3), 7: (1, 2, 3), 6: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_dom_positions(self):
        self.det._parse_doms()
        assert np.allclose([1.49992331, 1.51893187, 1.44185513],
                           self.det.dom_positions[1])
        assert np.allclose([2.49992331, 2.51893187, 2.44185513],
                           self.det.dom_positions[2])
        assert np.allclose([3.49992331, 3.51893187, 3.44185513],
                           self.det.dom_positions[3])

    def test_xy_positions(self):
        self.det._parse_doms()
        assert len(self.det.xy_positions) == 1
        assert np.allclose([1.49992331, 1.51893187], self.det.xy_positions[0])

    def test_correct_number_of_pmts(self):
        self.det._parse_doms()
        assert 9 == len(self.det.pmts)

    def test_pmt_attributes(self):
        self.det._parse_doms()
        assert (1, 2, 3, 4, 5, 6, 7, 8, 9) == tuple(self.det.pmts.pmt_id)
        assert np.allclose([1.1, 1.4, 1.7, 2.1, 2.4, 2.7, 3.1, 3.4, 3.7],
                           self.det.pmts.pos_x)
        assert np.allclose((1.7, 1.8, 1.9), self.det.pmts.pos[2])
        assert np.allclose((0.1, 0.2, -1.3), self.det.pmts.dir[8])

    def test_pmt_index_by_omkey(self):
        self.det._parse_doms()
        assert 5 == self.det._pmt_index_by_omkey[(1, 2, 2)]
        assert 0 == self.det._pmt_index_by_omkey[(1, 1, 0)]
        assert 4 == self.det._pmt_index_by_omkey[(1, 2, 1)]
        assert 1 == self.det._pmt_index_by_omkey[(1, 1, 1)]

    def test_pmt_index_by_pmt_id(self):
        self.det._parse_doms()
        assert 0 == self.det._pmt_index_by_pmt_id[1]

    def test_pmt_with_id_returns_correct_omkeys(self):
        self.det._parse_doms()
        pmt = self.det.pmt_with_id(1)
        assert (1, 1, 0) == (pmt.du, pmt.floor, pmt.channel_id)
        pmt = self.det.pmt_with_id(5)
        assert (1, 2, 1) == (pmt.du, pmt.floor, pmt.channel_id)

    def test_pmt_with_id_returns_correct_omkeys_with_mixed_pmt_ids(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        pmt = self.det.pmt_with_id(73)
        assert (1, 2, 1) == (pmt.du, pmt.floor, pmt.channel_id)
        pmt = self.det.pmt_with_id(81)
        assert (1, 1, 1) == (pmt.du, pmt.floor, pmt.channel_id)

    def test_pmt_with_id_raises_exception_for_invalid_id(self):
        self.det._parse_doms()
        with self.assertRaises(KeyError):
            self.det.pmt_with_id(100)

    def test_get_pmt(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        pmt = self.det.get_pmt(7, 2)
        assert (1, 2, 2) == (pmt.du, pmt.floor, pmt.channel_id)

    def test_xy_pos(self):
        self.det._parse_doms()
        xy = self.det.xy_positions
        assert xy is not None

    def test_ascii(self):
        detx_string = "\n".join((
            "1 3",
            "1 1 1 3",
            " 1 1.1 1.2 1.3 1.1 2.1 3.1 10.0",
            " 2 1.4 1.5 1.6 4.1 5.1 6.1 20.0",
            " 3 1.7 1.8 1.9 7.1 8.1 9.1 30.0",
            "2 1 2 3",
            " 4 2.1 2.2 2.3 1.2 2.2 3.2 40.0",
            " 5 2.4 2.5 2.6 4.2 5.2 6.2 50.0",
            " 6 2.7 2.8 2.9 7.2 8.2 9.2 60.0",
            "3 1 3 3",
            " 7 3.1 3.2 3.3 1.3 2.3 3.3 70.0",
            " 8 3.4 3.5 3.6 4.3 5.3 6.3 80.0",
            " 9 3.7 3.8 3.9 7.3 8.3 9.3 90.0\n",
        ))
        detx_fob = StringIO(detx_string)

        self.det = Detector()
        self.det._det_file = detx_fob
        self.det._parse_header()
        self.det._parse_doms()
        assert detx_string == self.det.ascii

    def test_ascii_with_mixed_dom_ids(self):
        detx_string = "\n".join((
            "1 3",
            "8 1 1 3",
            " 1 1.1 1.2 1.3 1.1 2.1 3.1 10.0",
            " 2 1.4 1.5 1.6 4.1 5.1 6.1 20.0",
            " 3 1.7 1.8 1.9 7.1 8.1 9.1 30.0",
            "4 1 2 3",
            " 4 2.1 2.2 2.3 1.2 2.2 3.2 40.0",
            " 5 2.4 2.5 2.6 4.2 5.2 6.2 50.0",
            " 6 2.7 2.8 2.9 7.2 8.2 9.2 60.0",
            "9 1 3 3",
            " 7 3.1 3.2 3.3 1.3 2.3 3.3 70.0",
            " 8 3.4 3.5 3.6 4.3 5.3 6.3 80.0",
            " 9 3.7 3.8 3.9 7.3 8.3 9.3 90.0\n",
        ))
        detx_fobj = StringIO(detx_string)

        self.det = Detector()
        self.det._det_file = detx_fobj
        self.det._parse_header()
        self.det._parse_doms()
        assert detx_string == self.det.ascii

    def test_detx_format_version_1(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v1.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 'v1' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detx_v1_is_the_same_ascii(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v1.detx'))
        with open(join(TEST_DATA_DIR, 'detx_v1.detx'), 'r') as fobj:
            assert fobj.read() == det.ascii

    def test_detx_format_version_2(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v2.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 256500.0 == det.utm_info.easting
        assert 4743000.0 == det.utm_info.northing
        assert 'WGS84' == det.utm_info.ellipsoid
        assert '32N' == det.utm_info.grid
        assert -2425.0 == det.utm_info.z
        assert 1500000000.1 == det.valid_from
        assert 9999999999.0 == det.valid_until
        assert 'v2' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detx_v2_is_the_same_ascii(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v2.detx'))
        with open(join(TEST_DATA_DIR, 'detx_v2.detx'), 'r') as fobj:
            assert fobj.read() == det.ascii

    def test_detx_format_version_3(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v3.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 256500.0 == det.utm_info.easting
        assert 4743000.0 == det.utm_info.northing
        assert 'WGS84' == det.utm_info.ellipsoid
        assert '32N' == det.utm_info.grid
        assert -2425.0 == det.utm_info.z
        assert 1500000000.1 == det.valid_from
        assert 9999999999.0 == det.valid_until
        assert 'v3' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detector_repr(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v3.detx'))
        assert "Detector id: '23', n_doms: 6, dus: [1, 2]" == repr(det)

    def test_detx_format_version_3_with_whitespace(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v3_whitespace.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 256500.0 == det.utm_info.easting
        assert 4743000.0 == det.utm_info.northing
        assert 'WGS84' == det.utm_info.ellipsoid
        assert '32N' == det.utm_info.grid
        assert -2425.0 == det.utm_info.z
        assert 1500000000.1 == det.valid_from
        assert 9999999999.0 == det.valid_until
        assert 'v3' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detx_format_comments(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v1.detx'))
        assert len(det.comments) == 0

        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v2.detx'))
        assert len(det.comments) == 0

        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v3.detx'))
        assert len(det.comments) == 2
        assert " a comment line" == det.comments[0]
        assert " another comment line starting with '#'" == det.comments[1]

    def test_comments_are_written(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v3.detx'))
        det.add_comment("foo")
        assert 3 == len(det.comments)
        assert det.comments[2] == "foo"
        assert "# foo" == det.ascii.splitlines()[2]

    def test_detx_v3_is_the_same_ascii(self):
        det = Detector(filename=join(TEST_DATA_DIR, 'detx_v3.detx'))
        with open(join(TEST_DATA_DIR, 'detx_v3.detx'), 'r') as fobj:
            assert fobj.read() == det.ascii

    def test_translate_detector(self):
        self.det._parse_doms()
        t_vec = np.array([1, 2, 3])
        orig_pos = self.det.pmts.pos.copy()
        self.det.translate_detector(t_vec)
        assert np.allclose(orig_pos + t_vec, self.det.pmts.pos)

    def test_translate_detector_updates_xy_positions(self):
        self.det._parse_doms()
        t_vec = np.array([1, 2, 3])
        orig_xy_pos = self.det.xy_positions.copy()
        self.det.translate_detector(t_vec)
        assert np.allclose(orig_xy_pos + t_vec[:2], self.det.xy_positions)

    def test_translate_detector_updates_dom_positions(self):
        self.det._parse_doms()
        t_vec = np.array([1, 2, 3])
        orig_dom_pos = deepcopy(self.det.dom_positions)
        self.det.translate_detector(t_vec)
        for dom_id, pos in self.det.dom_positions.items():
            assert np.allclose(orig_dom_pos[dom_id] + t_vec, pos)

    def test_rotate_dom_by_yaw(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()
        # here, only one PMT is checked
        dom_id = 1
        heading = 23
        channel_id = 0
        pmt_dir = det.pmts[det.pmts.dom_id == dom_id].dir[channel_id].copy()
        pmt_dir_rot = qrot_yaw(pmt_dir, heading)
        det.rotate_dom_by_yaw(dom_id, heading)
        assert np.allclose(
            pmt_dir_rot, det.pmts[det.pmts.dom_id == dom_id].dir[channel_id]
        )
        assert np.allclose([0.92050485, 0.39073113, 0],
                           det.pmts[det.pmts.dom_id == dom_id].pos[channel_id])

    def test_rotate_dom_set_by_step_by_360_degrees(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()
        dom_id = 1
        channel_id = 0
        pmt_dir = det.pmts[det.pmts.dom_id == dom_id].dir[channel_id].copy()
        pmt_pos = det.pmts[det.pmts.dom_id == dom_id].pos[channel_id].copy()
        for i in range(36):
            det.rotate_dom_by_yaw(dom_id, 10)
        pmt_dir_rot = det.pmts[det.pmts.dom_id == dom_id].dir[channel_id]
        assert np.allclose(pmt_dir, pmt_dir_rot)
        pmt_pos_rot = det.pmts[det.pmts.dom_id == dom_id].pos[channel_id]
        assert np.allclose(pmt_pos, pmt_pos_rot)

    def test_rotate_du_by_yaw_step_by_step_360_degrees(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()
        du = 2
        pmt_dir = det.pmts[det.pmts.du == du].dir.copy()
        pmt_pos = det.pmts[det.pmts.du == du].pos.copy()
        pmt_dir_other_dus = det.pmts[det.pmts.du != du].dir.copy()
        pmt_pos_other_dus = det.pmts[det.pmts.du != du].pos.copy()
        for i in range(36):
            det.rotate_du_by_yaw(du, 10)
        pmt_dir_rot = det.pmts[det.pmts.du == du].dir
        pmt_pos_rot = det.pmts[det.pmts.du == du].pos
        assert np.allclose(pmt_dir, pmt_dir_rot)
        assert np.allclose(pmt_pos, pmt_pos_rot)
        assert np.allclose(pmt_dir_other_dus, det.pmts[det.pmts.du != du].dir)
        assert np.allclose(pmt_pos_other_dus, det.pmts[det.pmts.du != du].pos)

    def test_rescale_detector(self):
        self.det._parse_doms()
        dom_positions = deepcopy(self.det.dom_positions)
        scale_factor = 2
        self.det.rescale(scale_factor)
        for dom_id, dom_pos in self.det.dom_positions.items():
            assert np.allclose(dom_pos, dom_positions[dom_id] * scale_factor)



#!/usr/bin/env python
# Filename: srv.py
"""
The KM3srv tornado webserver.

"""
from __future__ import absolute_import, print_function, division

try:
    import tornado
except ImportError:
    print(
        "Please 'pip install tornado websocket-client' "
        "to use the km3srv package"
    )
    exit(1)

import tornado.ioloop
import tornado.web
import tornado.websocket
from tornado.options import define, options

import os
import threading
import re
import subprocess
import math

from time import sleep

import pandas as pd
import websocket

from .calib import Calibration
from .config import Config
from .dataclasses import Table
from .tools import token_urlsafe
from .logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)

VERSION = '0.0.1'

define(
    "ip",
    default="0.0.0.0",
    type=str,
    help="The WAN IP of this machine. You can use 127 for local tests."
)
define(
    "port",
    default="8088",
    type=int,
    help="The KM3srv server will be available on this port."
)
define(
    "data",
    default=os.path.expanduser("~/km3net/data"),
    type=str,
    help="Path to the data files."
)

RBA_URL = Config().rba_url


class ClientManager(object):
    """Manage km3srv clients.
    """

    def __init__(self):
        self._clients = {}
        threading.Thread(target=self.heartbeat).start()

    def add(self, client):
        token = token_urlsafe(3)
        self._clients[token] = client
        log.info("New client with token '{0}' registered.".format(token))
        return token

    def remove(self, token):
        del self._clients[token]
        print("Client with token '{0}' removed.".format(token))

    def heartbeat(self, interval=30):
        while True:
            log.info("Pinging {0} clients.".format(len(self._clients)))
            print(self._clients)
            for client in self._clients.values():
                print(client)
                client.message("Ping.")
            sleep(interval)

    def broadcast_status(self):
        # status = {
        #    n_clients : len(self._clients)
        # }
        self.broadcast(
            "Number of connected clients: {0}.".format(len(self._clients))
        )

    def message_to(self, token, data, kind):
        message = pd.io.json.dumps({'kind': kind, 'data': data})
        print("Sent {0} bytes.".format(len(message)))
        self.raw_message_to(token, message)

    def raw_message_to(self, token, message):
        """Convert message to JSON and send it to the client with token"""
        if token not in self._clients:
            log.critical("Client with token '{0}' not found!".format(token))
            return
        client = self._clients[token]
        try:
            client.write_message(message)
        except (AttributeError, tornado.websocket.WebSocketClosedError):
            log.error("Lost connection to client '{0}'".format(client))
        else:
            print("Sent {0} bytes.".format(len(message)))

    def broadcast(self, data, kind="info"):
        log.info("Broatcasting to {0} clients.".format(len(self._clients)))
        for token in self._clients.keys():
            self.message_to(token, data, kind)


class MessageProvider(tornado.websocket.WebSocketHandler):
    def __init__(self, *args, **kwargs):
        self.client_manager = kwargs.pop('client_manager')
        super(MessageProvider, self).__init__(*args, **kwargs)

    def on_message(self, message):
        log.warning("Client said: {0}".format(message))
        try:
            token = pd.io.json.loads(message)['token']
        except (ValueError, KeyError):
            log.error("Invalid JSON received: {0}".format(message))
        else:
            self.client_manager.raw_message_to(token, message)


class EchoWebSocket(tornado.websocket.WebSocketHandler):
    """An echo handler for client/server messaging and debugging"""

    def __init__(self, *args, **kwargs):
        log.warning("EchoWebSocket created!")
        self.client_manager = kwargs.pop('client_manager')
        self.data_path = kwargs.pop('data_path')
        self._status = kwargs.pop('server_status')
        self._lock = kwargs.pop('lock')
        self._token = self.client_manager.add(self)
        super(EchoWebSocket, self).__init__(*args, **kwargs)

    def check_origin(self, origin):
        return True

    def open(self):
        welcome = u"WebSocket opened. Welcome to KM3srv {0}!".format(VERSION)
        self.message(welcome)
        self.message(self._token, 'token')
        self.message(self.status, 'status')
        self.client_manager.broadcast_status()

    def on_close(self):
        self.client_manager.remove(self._token)
        print("WebSocket closed, client removed.")

    def on_message(self, message):
        # self.message(u"Client said '{0}'".format(message))
        print("Client said: {0}".format(message))
        if message.startswith('event'):
            p = re.compile(r'event/(\d+)/(\d+)/(\d+)')
            try:
                det_id, run_id, event_id = re.search(p, message).groups()
            except AttributeError:
                self.message("Syntax error, try event/DET_ID/RUN_ID/EVENT")
            else:
                threading.Thread(
                    target=self.get_event,
                    args=(int(det_id), int(run_id), int(event_id))
                ).start()

    def get_event(self, det_id, run_id, event_id):
        det_dir_name = "KM3NeT_{0:08d}".format(det_id)
        det_dir = os.path.join(self.data_path, 'sea', det_dir_name)
        sub_dir = str(int(math.floor(run_id / 1000)))
        data_dir = os.path.join(det_dir, sub_dir)

        basename = "{0}_{1:08d}".format(det_dir_name, run_id, event_id)
        h5filename = basename + ".h5"
        rootfilename = basename + ".root"

        irods_path = os.path.join(
            "/in2p3/km3net/data/raw/sea", det_dir_name, sub_dir, rootfilename
        )

        h5filepath = os.path.join(data_dir, h5filename)
        h5filepath_tmp = h5filepath + '.tmp'
        rootfilepath = os.path.join(data_dir, rootfilename)

        self.message("Looking for {0}".format(basename))
        print("Request for event {0} in {1}".format(event_id, h5filename))
        print("Detector dir: {0}".format(det_dir))
        print("Data dir: {0}".format(data_dir))

        if os.path.exists(h5filepath_tmp):
            self.message("File is currently being process. Waiting...")

        while os.path.exists(h5filepath_tmp):
            sleep(3)

        if not os.path.exists(h5filepath):
            self.status = 'busy'
            print("Creating {0}".format(h5filepath_tmp))
            subprocess.call(['mkdir', '-p', data_dir])
            subprocess.call(['touch', h5filepath_tmp])
            if not os.path.exists(rootfilepath):
                self.message("No HDF5 file found, downloading ROOT file.")
                print("Retrieve {0}".format(irods_path))
                status = subprocess.call(['iget', irods_path, data_dir])
                if status == 0:
                    self.message("Successfully downloaded data.")
                else:
                    self.message("There was a problem downloading the data.")
                    return
            status = subprocess.call([
                'km3pipe', 'jpptohdf5', '-i', rootfilepath, '-o',
                h5filepath_tmp
            ])
            if status == 0:
                self.message("Successfully converted data.")
                os.rename(h5filepath_tmp, h5filepath)
            else:
                self.message("There was a problem converting the data.")

        if os.path.exists(rootfilepath):
            log.warning("Removing ROOT file {0}".format(rootfilepath))
            os.remove(rootfilepath)

        hits = pd.read_hdf(h5filepath, 'hits', mode='r')
        snapshot_hits = hits[(hits['event_id'] == event_id)].copy()
        triggered_hits = hits[(hits['event_id'] == event_id) &
                              (hits['triggered'] == True)]    # noqa
        self.message(
            "Det ID: {0} Run ID: {1} Event ID: {2} - "
            "Snapshot hits: {3}, Triggered hits: {4}".format(
                det_id, run_id, event_id, len(snapshot_hits),
                len(triggered_hits)
            )
        )
        cal = Calibration(det_id=det_id)
        cal.apply(snapshot_hits)

        event = {
            "hits": {
                'pos': [
                    tuple(x) for x in snapshot_hits[['x', 'y', 'z']].values
                ],
                'time': list(snapshot_hits['time']),
                'tot': list(snapshot_hits['tot']),
            }
        }

        self.message(event, "event")
        self.status = 'ready'

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value
        # self.broadcast(self.status, kind='status')

    def message(self, data, kind="info"):
        """Convert message to json and send it to the clients"""
        message = pd.io.json.dumps({'kind': kind, 'data': data})
        print("Sent {0} bytes.".format(len(message)))
        self.write_message(message)


def srv_event(token, hits, url=RBA_URL):
    """Serve event to RainbowAlga"""

    if url is None:
        log.error("Please provide a valid RainbowAlga URL.")
        return

    ws_url = url + '/message'

    if isinstance(hits, pd.core.frame.DataFrame):
        pos = [tuple(x) for x in hits[['x', 'y', 'z']].values]
        time = list(hits['time'])
        tot = list(hits['tot'])
    elif isinstance(hits, Table):
        pos = list(zip(hits.pos_x, hits.pos_y, hits.pos_z))
        time = list(hits.time)
        tot = list(hits.tot)
    else:
        log.error(
            "No calibration information found in hits (type: {0})".format(
                type(hits)
            )
        )
        return

    event = {
        "hits": {
            'pos': pos,
            'time': time,
            'tot': tot,
        }
    }

    srv_data(ws_url, token, event, 'event')


def srv_data(url, token, data, kind):
    """Serve data to RainbowAlga"""
    ws = websocket.create_connection(url)
    message = {'token': token, 'data': data, 'kind': kind}
    ws.send(pd.io.json.dumps(message))
    ws.close()


def main():
    root = os.path.dirname(__file__)

    options.parse_command_line()

    ip = options.ip
    port = int(options.port)
    data_path = options.data
    server_status = 'ready'
    lock = threading.Lock()
    client_manager = ClientManager()

    print("Starting KM3srv.")
    print("Running on {0}:{1}".format(ip, port))
    print("Data path: {0}".format(data_path))

    settings = {
        'debug': True,
        'static_path': os.path.join(root, 'static'),
        'template_path': os.path.join(root, 'static/templates'),
    }

    application = tornado.web.Application([
        (
            r"/test", EchoWebSocket,
            {
                'client_manager': client_manager,
                'server_status': server_status,
                'data_path': data_path,
                'lock': lock
            }
        ),
        (r"/message", MessageProvider,
         {
             'client_manager': client_manager
         }),
    ], **settings)

    try:
        application.listen(port)
        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        print("Stopping tornado...")
        tornado.ioloop.IOLoop.instance().stop()


if __name__ == "__main__":
    main()



# Filename: test_stats.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

import numpy as np
from numpy.random import RandomState
from pandas import DataFrame
import pytest
from scipy import stats

from km3pipe.testing import TestCase
from km3pipe.stats import (
    loguniform, rv_kde, mad, mad_std, drop_zero_variance, param_names, perc,
    resample_1d, bootstrap_params, param_describe, bootstrap_fit, hist2d,
    bincenters
)

__author__ = ["Tamas Gal", "Moritz Lotze"]
__copyright__ = "Copyright 2016, KM3Pipe devs and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__maintainer__ = ["Tamas Gal", "Moritz Lotze"]
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestLogUniform(TestCase):
    def setUp(self):
        np.random.seed(1234)

    def test_rvs(self):
        lo, hi = 0.1, 10
        dist = loguniform(low=lo, high=hi, base=10)
        r = dist.rvs(size=100)
        assert r.shape == (100, )
        assert np.all(r <= hi)
        assert np.all(r >= lo)
        dist = loguniform(low=lo, high=hi, base=2)
        r2 = dist.rvs(size=500)
        assert r2.shape == (500, )
        assert np.all(r2 <= hi)
        assert np.all(r2 >= lo)


class TestRvKde(TestCase):
    def setUp(self):
        self.data1d = np.random.normal(size=20)
        self.data = self.data1d[:, np.newaxis]

    def test_shape(self):
        with pytest.raises(ValueError):
            rv = rv_kde(self.data1d)
        rv = rv_kde(self.data)
        samp = rv.rvs(size=10)
        assert len(samp) == 10
        assert samp.ndim == 2
        assert samp.shape == (10, 1)
        samp = rv.rvs(size=(10, ))
        assert len(samp) == 10
        assert samp.ndim == 2
        assert samp.shape == (10, 1)

    def test_2d(self):
        d2d = np.column_stack([
            np.random.normal(size=20),
            np.random.normal(size=20),
        ])
        rv = rv_kde(d2d)
        samp = rv.rvs(size=10)
        assert len(samp) == 10
        assert samp.ndim == 2
        assert samp.shape == (10, 2)

    def test_bw_methods(self):
        rv = rv_kde(
            self.data,
            bw_method='scott',
        )
        assert rv is not None

    def test_bw_statsmodels(self):
        rv = rv_kde(
            self.data,
            bw_statsmodels=True,
        )
        assert rv is not None

    def test_bw_statsmodels_ignores_method(self):
        rv1 = rv_kde(
            self.data,
            bw_statsmodels=True,
        )
        rv2 = rv_kde(
            self.data,
            bw_statsmodels=True,
            bw_method='scott',
        )
        rv3 = rv_kde(
            self.data,
            bw_statsmodels=True,
            bw_method='silverman',
        )
        assert rv2 is not None
        assert rv1 is not None
        assert rv1.bandwidth == rv2.bandwidth
        assert rv1.bandwidth == rv3.bandwidth
        assert rv2.bandwidth == rv3.bandwidth


class TestMAD(TestCase):
    def test_wiki(self):
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        assert np.allclose(mad(arr), 1)

    def test_normal(self):
        np.random.seed(42)
        sample = np.random.normal(0, 1, 1000)
        assert np.allclose(mad_std(sample), 1, atol=.1)


class TestPerc(TestCase):
    def test_settings(self):
        arr = np.array([
            -2.394,
            0.293,
            0.371,
            0.384,
            1.246,
        ])
        assert np.allclose(perc(arr), [-2.1253, 1.1598])
        assert np.allclose(
            perc(arr, interpolation='nearest'), [arr[0], arr[-1]]
        )


class TestVariance(TestCase):
    def test_easy(self):
        df = DataFrame({'a': [1, 2, 3], 'b': [1, 1, 1], 'c': [4, 4, 3]})
        df2 = drop_zero_variance(df)
        assert df.shape == (3, 3)
        assert df2.shape == (3, 2)
        assert 'b' not in df2.columns
        assert 'b' in df.columns


class TestResample(TestCase):
    def test_shape(self):
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        r = resample_1d(arr)
        assert arr.shape == r.shape

    def test_seed(self):
        seed = 42
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        r1 = resample_1d(arr, random_state=RandomState(seed=seed))
        r2 = resample_1d(arr, random_state=RandomState(seed=seed))
        assert np.allclose(r1, r2)
        assert np.allclose([9, 2, 4, 9, 2, 4, 4], r1)


class TestBootstrapFit(TestCase):
    def setUp(self):
        self.seed = 42

    def test_raise_on_instance(self):
        with pytest.raises(TypeError):
            n = param_names(stats.norm())
            assert n is not None

    def test_names(self):
        assert param_names(stats.norm) == ['loc', 'scale']
        assert param_names(stats.lognorm) == ['loc', 'scale', 's']

    def test_raw(self):
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        pars = bootstrap_params(
            stats.norm, arr, n_iter=500, random_state=RandomState(self.seed)
        )
        assert np.allclose(
            np.mean(pars, axis=0), [np.mean(arr), np.std(arr)], atol=.5
        )

    def test_param_describe(self):
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        pars = bootstrap_params(
            stats.norm, arr, n_iter=500, random_state=RandomState(self.seed)
        )
        desc = param_describe(pars)
        assert desc is not None
        assert len(desc) == 3
        assert desc[0].shape[0] == 2

    def test_full(self):
        arr = np.array([1, 1, 2, 2, 4, 6, 9])
        fits = bootstrap_fit(
            stats.norm, arr, n_iter=500, random_state=RandomState(self.seed)
        )
        assert np.allclose(fits['mean'], [np.mean(arr), np.std(arr)], atol=.5)


class TestHist2D(TestCase):
    def setUp(self):
        np.random.seed(42)
        self.sample = np.random.normal(0, 1, (100, 2))
        self.bins = ([-2, 0, 2], [-2, 0, 2])

    def test_pdf(self):
        hist = np.histogram2d(
            self.sample[:, 0], self.sample[:, 1], normed=True, bins=self.bins
        )
        h = hist2d(hist)
        assert h.H.shape == (2, 2)
        assert h.H_pad.shape == (4, 4)
        f = h.pdf([[0, 0]])
        assert f is not None

    def test_norm(self):
        hist1 = np.histogram2d(
            self.sample[:, 0], self.sample[:, 1], normed=True, bins=self.bins
        )
        hist2 = np.histogram2d(
            self.sample[:, 0], self.sample[:, 1], normed=False, bins=self.bins
        )
        h1 = hist2d(hist1)
        h2 = hist2d(hist2)

        print(h1.area)
        print(h1.H)
        print(h1.H / h1.area)
        print((h1.H / h1.area).sum())
        print(h1.H.sum())

        h1s = h1.H.sum()
        assert np.allclose(h1s, h1.H.sum())

        assert np.allclose(h1.H.sum(), h2.H.sum())

        assert np.allclose(h1.integral, 1)
        assert np.allclose(h2.integral, 1)


class TestBins(TestCase):
    def test_binlims(self):
        bins = np.linspace(0, 20, 21)
        assert bincenters(bins).shape[0] == bins.shape[0] - 1



# Filename: test_root.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212

from km3pipe.testing import TestCase, patch, Mock, surrogate
from km3pipe.io.root import (
    open_rfile, get_hist, get_hist2d, get_hist3d, interpol_hist2d
)

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestOpenRfile(TestCase):
    @surrogate('rootpy.io.root_open')
    @patch('rootpy.io.root_open')
    def test_call(self, root_open_mock):
        open_rfile('a.root')
        root_open_mock.assert_called_with('a.root', mode='r')


class TestGetHist(TestCase):
    @surrogate('root_numpy.hist2array')
    @patch('km3pipe.io.root.open_rfile')
    @patch('root_numpy.hist2array')
    def test_call(self, hist2array_mock, open_rfile_mock):
        get_hist('a.root', 'histname')
        hist2array_mock.assert_called_once()


class TestGetHist2d(TestCase):
    @surrogate('root_numpy.hist2array')
    @patch('km3pipe.io.root.open_rfile')
    @patch('root_numpy.hist2array')
    def test_call(self, hist2array_mock, open_rfile_mock):
        get_hist2d('a.root', 'histname')
        hist2array_mock.assert_called_once()


class TestGetHist3d(TestCase):
    @surrogate('root_numpy.hist2array')
    @patch('km3pipe.io.root.open_rfile')
    @patch('root_numpy.hist2array')
    def test_call(self, hist2array_mock, open_rfile_mock):
        get_hist3d('a.root', 'histname')
        hist2array_mock.assert_called_once()



# Filename: test_aanet.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from os.path import join, dirname

import numpy as np

from km3pipe.testing import TestCase, patch, Mock, skip
from km3pipe.io.aanet import AanetPump, MetaParser

import sys
sys.modules['ROOT'] = Mock()
sys.modules['aa'] = Mock()

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

TEST_DATA_DIR = join(dirname(__file__), "../../kp-data/test_data")


class TestAanetPump(TestCase):
    def test_init_raises_valueerror_if_no_filename_given(self):
        with self.assertRaises(TypeError):
            AanetPump()

    @skip
    def test_init_with_filename(self):
        filename = 'a'
        p = AanetPump(filename=filename)
        assert filename in p.filenames


class TestMetaParser(TestCase):
    def test_init(self):
        MetaParser()

    def test_parse_string_of_single_entry(self):
        string = b"A 123\nA 1.2.3\nA KM3NET\nA a\nA b\nA Linux"
        mp = MetaParser()
        mp.parse_string(string)

        assert 1 == len(mp.meta)

        assert b'A' == mp.meta[0]['application_name']
        assert b'123' == mp.meta[0]['revision']
        assert b'A a\nA b' == mp.meta[0]['command']

    def test_parse_string_of_multiple_entries(self):
        string = (
            b"A 123\nA 1.2.3\nA KM3NET\nA a\nA b\nA Linux\n"
            b"B 456\nB 4.5.6\nB KM3NET\nB c\nB Linux"
        )
        mp = MetaParser()
        mp.parse_string(string)

        assert 2 == len(mp.meta)

        assert b'A' == mp.meta[0]['application_name']
        assert b'123' == mp.meta[0]['revision']
        assert b'1.2.3' == mp.meta[0]['root_version']
        assert b'A a\nA b' == mp.meta[0]['command']

        assert b'B' == mp.meta[1]['application_name']
        assert b'456' == mp.meta[1]['revision']
        assert b'4.5.6' == mp.meta[1]['root_version']
        assert b'B c' == mp.meta[1]['command']

    def test_parse_testfile(self):
        fname = join(TEST_DATA_DIR, 'jprintmeta.log')
        with open(fname, 'rb') as fobj:
            string = fobj.read()
        mp = MetaParser()
        mp.parse_string(string)
        assert 7 == len(mp.meta)

        assert b'JEvt' == mp.meta[0]['application_name']
        assert b'9912' == mp.meta[0]['revision']
        assert b'5.34/23' == mp.meta[0]['root_version']
        assert mp.meta[0]['command'].startswith(b'JEvt /pbs/throng/km3net')
        assert mp.meta[0]['command'].endswith(b'2 --!')

        assert b'JEnergy' == mp.meta[1]['application_name']
        assert b'9912' == mp.meta[1]['revision']
        assert b'5.34/23' == mp.meta[1]['root_version']
        assert mp.meta[1]['command'].startswith(b'JEnergy /pbs/throng/km3net')
        assert mp.meta[1]['command'].endswith(b'1 --!')

        assert b'JTriggerEfficiency' == mp.meta[-1]['application_name']
        assert b'8519' == mp.meta[-1]['revision']
        assert b'5.34/23' == mp.meta[-1]['root_version']
        assert mp.meta[-1]['command'].startswith(
            b'JTriggerEfficiency /pbs/throng/km3net'
        )
        assert mp.meta[-1]['command'].endswith(b'326 --!')

    def test_get_table(self):
        string = (
            b"A 123\nA 1.2.3\nA KM3NET\nA a\nA b\nA Linux\n"
            b"B 456\nB 4.5.6\nB KM3NET\nB c\nB Linux"
        )
        mp = MetaParser()
        mp.parse_string(string)
        tab = mp.get_table()
        assert 2 == len(tab)
        assert '/meta' == tab.h5loc
        assert tab.h5singleton
        assert b'B c' == tab['command'][1]
        assert np.dtype('S7') == tab['command'].dtype

    def test_empty_string(self):
        mp = MetaParser()
        mp.parse_string('')
        assert mp.get_table() is None



# Filename: hdf5.py
# pylint: disable=C0103,R0903,C901
# vim:set ts=4 sts=4 sw=4 et:
"""
Read and write KM3NeT-formatted HDF5 files.

"""
from __future__ import absolute_import, print_function, division

from collections import OrderedDict, defaultdict, namedtuple
import os.path
import warnings

import numpy as np
import tables as tb

import km3pipe as kp
from km3pipe.core import Pump, Module, Blob
from km3pipe.dataclasses import Table, NDArray, DEFAULT_H5LOC
from km3pipe.logger import get_logger
from km3pipe.tools import decamelise, camelise, split, istype, get_jpp_revision

log = get_logger(__name__)    # pylint: disable=C0103

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

FORMAT_VERSION = np.string_('5.1')
MINIMUM_FORMAT_VERSION = np.string_('4.1')


class H5VersionError(Exception):
    pass


def check_version(h5file):
    try:
        version = np.string_(h5file.root._v_attrs.format_version)
    except AttributeError:
        log.error(
            "Could not determine HDF5 format version: '%s'."
            "You may encounter unexpected errors! Good luck..." %
            h5file.filename
        )
        return
    if split(version, int, np.string_('.')) < \
            split(MINIMUM_FORMAT_VERSION, int, np.string_('.')):
        raise H5VersionError(
            "HDF5 format version {0} or newer required!\n"
            "'{1}' has HDF5 format version {2}.".format(
                MINIMUM_FORMAT_VERSION.decode("utf-8"), h5file.filename,
                version.decode("utf-8")
            )
        )


class HDF5Header(object):
    """Wrapper class for the `/raw_header` table in KM3HDF5"""

    def __init__(self, data):
        self._data = data
        self._set_attributes()

    def _set_attributes(self):
        """Traverse the internal dictionary and set the getters"""
        for parameter, data in self._data.items():
            if isinstance(data, dict) or isinstance(data, OrderedDict):
                field_names, field_values = zip(*data.items())
                sorted_indices = np.argsort(field_names)
                attr = namedtuple(
                    parameter,
                    [field_names[i] for i in sorted_indices]
                )
                setattr(
                    self, parameter,
                    attr(*[field_values[i] for i in sorted_indices])
                )
            else:
                setattr(self, parameter, data)

    @classmethod
    def from_table(cls, table):
        data = OrderedDict()
        for i in range(len(table)):
            parameter = table['parameter'][i]
            field_names = table['field_names'][i].split(' ')
            field_values = table['field_values'][i].split(' ')
            if field_values == ['']:
                log.info(
                    "No value for parameter '{}'! Skipping...".
                    format(parameter)
                )
                continue
            dtypes = table['dtype'][i]
            dtyped_values = []
            for dtype, value in zip(dtypes.split(' '), field_values):
                if dtype.startswith('a'):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=' ')[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)

    @classmethod
    def from_hdf5(cls, filename):
        with tb.open_file(filename, 'r') as f:
            table = f.get_node('/raw_header')
            return cls.from_pytable(table)

    @classmethod
    def from_pytable(cls, table):
        data = OrderedDict()
        for row in table:
            parameter = row['parameter'].decode()
            field_names = row['field_names'].decode().split(' ')
            field_values = row['field_values'].decode().split(' ')
            if field_values == ['']:
                log.info(
                    "No value for parameter '{}'! Skipping...".
                    format(parameter)
                )
                continue
            dtypes = row['dtype'].decode()
            dtyped_values = []
            for dtype, value in zip(dtypes.split(' '), field_values):
                if dtype.startswith('a'):
                    dtyped_values.append(value)
                else:
                    value = np.fromstring(value, dtype=dtype, sep=' ')[0]
                    dtyped_values.append(value)
            data[parameter] = OrderedDict(zip(field_names, dtyped_values))
        return cls(data)


class HDF5IndexTable(object):
    def __init__(self, h5loc):
        self.h5loc = h5loc
        self._data = defaultdict(list)
        self._index = 0

    def append(self, n_items):
        self._data['indices'].append(self._index)
        self._data['n_items'].append(n_items)
        self._index += n_items

    @property
    def data(self):
        return self._data


class HDF5Sink(Module):
    """Write KM3NeT-formatted HDF5 files, event-by-event.

    The data can be a ``kp.Table``, a numpy structured array,
    a pandas DataFrame, or a simple scalar.

    The name of the corresponding H5 table is the decamelised
    blob-key, so values which are stored in the blob under `FooBar`
    will be written to `/foo_bar` in the HDF5 file.

    Parameters
    ----------
    filename: str, optional (default: 'dump.h5')
        Where to store the events.
    h5file: pytables.File instance, optional (default: None)
        Opened file to write to. This is mutually exclusive with filename.
    pytab_file_args: dict [optional]
        pass more arguments to the pytables File init
    n_rows_expected = int, optional [default: 10000]
    append: bool, optional [default: False]

    """

    def configure(self):
        self.filename = self.get('filename', default='dump.h5')
        self.ext_h5file = self.get('h5file')
        self.pytab_file_args = self.get('pytab_file_args', default=dict())
        self.file_mode = 'a' if self.get('append') else 'w'
        self.keep_open = self.get('keep_open')
        self.indices = {}    # to store HDF5IndexTables for each h5loc
        self._singletons_written = {}
        # magic 10000: this is the default of the "expectedrows" arg
        # from the tables.File.create_table() function
        # at least according to the docs
        # might be able to set to `None`, I don't know...
        self.n_rows_expected = self.get('n_rows_expected') or 10000
        self.index = 0

        if self.ext_h5file is not None:
            self.h5file = self.ext_h5file
        else:
            self.h5file = tb.open_file(
                self.filename,
                mode=self.file_mode,
                title="KM3NeT",
                **self.pytab_file_args
            )
        self.filters = tb.Filters(
            complevel=5, shuffle=True, fletcher32=True, complib='zlib'
        )
        self._tables = OrderedDict()
        self._ndarrays = OrderedDict()

    def _to_array(self, data, name=None):
        if data is None:
            return
        if np.isscalar(data):
            self.log.debug('toarray: is a scalar')
            return Table({
                name: np.asarray(data).reshape((1, ))
            },
                         h5loc='/misc/{}'.format(decamelise(name)),
                         name=name)
        if hasattr(data, 'len') and len(data) <= 0:    # a bit smelly ;)
            self.log.debug('toarray: data has no length')
            return
        # istype instead isinstance, to avoid heavy pandas import (hmmm...)
        if istype(data, 'DataFrame'):    # noqa
            self.log.debug('toarray: pandas dataframe')
            data = Table.from_dataframe(data)
        return data

    def _write_ndarray(self, arr):
        h5loc = arr.h5loc
        title = arr.title
        if h5loc not in self._ndarrays:
            loc, tabname = os.path.split(h5loc)
            ndarr = self.h5file.create_earray(
                loc,
                tabname,
                tb.Atom.from_dtype(arr.dtype),
                (0, ) + arr.shape[1:],
                title=title,
                filters=self.filters,
                createparents=True,
            )
            self._ndarrays[h5loc] = ndarr
        else:
            ndarr = self._ndarrays[h5loc]

        idx_table_h5loc = h5loc + '_indices'
        if idx_table_h5loc not in self.indices:
            self.indices[idx_table_h5loc] = HDF5IndexTable(idx_table_h5loc)
        idx_tab = self.indices[idx_table_h5loc]
        idx_tab.append(len(arr))

        ndarr.append(arr)

    def _write_table(self, h5loc, arr, title):
        level = len(h5loc.split('/'))

        if h5loc not in self._tables:
            dtype = arr.dtype
            loc, tabname = os.path.split(h5loc)
            self.log.debug(
                "h5loc '{}', Loc '{}', tabname '{}'".format(
                    h5loc, loc, tabname
                )
            )
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', tb.NaturalNameWarning)
                tab = self.h5file.create_table(
                    loc,
                    tabname,
                    description=dtype,
                    title=title,
                    filters=self.filters,
                    createparents=True,
                    expectedrows=self.n_rows_expected,
                )
            tab._v_attrs.datatype = title
            if (level < 5):
                self._tables[h5loc] = tab
        else:
            tab = self._tables[h5loc]

        h5_colnames = set(tab.colnames)
        tab_colnames = set(arr.dtype.names)
        if h5_colnames != tab_colnames:
            missing_cols = h5_colnames - tab_colnames
            if missing_cols:
                self.log.info(
                    "Missing columns in table, trying to append NaNs."
                )
                arr = arr.append_columns(
                    missing_cols, np.full((len(missing_cols), len(arr)),
                                          np.nan)
                )
        tab.append(arr)

        if (level < 4):
            tab.flush()

    def _write_separate_columns(self, where, obj, title):
        f = self.h5file
        loc, group_name = os.path.split(where)
        if where not in f:
            group = f.create_group(loc, group_name, title)
            group._v_attrs.datatype = title
        else:
            group = f.get_node(where)

        for col, (dt, _) in obj.dtype.fields.items():
            data = obj.__array__()[col]

            if col not in group:
                a = tb.Atom.from_dtype(dt)
                arr = f.create_earray(
                    group,
                    col,
                    a, (0, ),
                    col.capitalize(),
                    filters=self.filters
                )
            else:
                arr = getattr(group, col)
            arr.append(data)

        # create index table
        if where not in self.indices:
            self.indices[where] = HDF5IndexTable(where + '/_indices')
        idx_tab = self.indices[where]
        idx_tab.append(len(data))

    def _process_entry(self, key, entry):
        self.log.debug("Inspecting {}".format(key))
        if hasattr(
                entry, 'h5singleton'
        ) and entry.h5singleton and entry.h5loc in self._singletons_written:
            self.log.debug(
                "Skipping '%s' since it's a singleton and already written." %
                entry.h5loc
            )
            return
        if not hasattr(entry, 'h5loc'):
            self.log.debug("Ignoring '%s': no h5loc attribute" % key)
            return
        if isinstance(entry, NDArray):
            self._write_ndarray(entry)
            return entry
        try:
            title = entry.name
        except AttributeError:
            title = key

        if isinstance(entry, Table) and not entry.h5singleton:
            if 'group_id' not in entry:
                entry = entry.append_columns('group_id', self.index)

        self.log.debug("h5l: '{}', title '{}'".format(entry.h5loc, title))

        if hasattr(entry, 'split_h5') and entry.split_h5:
            self.log.debug("Writing into separate columns...")
            self._write_separate_columns(entry.h5loc, entry, title=title)
        else:
            self.log.debug("Writing into single Table...")
            self._write_table(entry.h5loc, entry, title=title)

        if hasattr(entry, 'h5singleton') and entry.h5singleton:
            self._singletons_written[entry.h5loc] = True

        return entry

    def process(self, blob):
        written_blob = Blob()
        for key, entry in sorted(blob.items()):
            data = self._process_entry(key, entry)
            if data is not None:
                written_blob[key] = data

        if 'GroupInfo' not in blob:
            gi = Table(
                {
                    'group_id': self.index,
                    'blob_length': len(written_blob)
                },
                h5loc='/group_info',
                name='Group Info',
            )
            self._process_entry('GroupInfo', gi)

        if not self.index % 1000:
            self.log.info('Flushing tables to disk...')
            for tab in self._tables.values():
                tab.flush()

        self.index += 1
        return blob

    def finish(self):
        self.h5file.root._v_attrs.km3pipe = np.string_(kp.__version__)
        self.h5file.root._v_attrs.pytables = np.string_(tb.__version__)
        self.h5file.root._v_attrs.jpp = np.string_(get_jpp_revision())
        self.h5file.root._v_attrs.format_version = np.string_(FORMAT_VERSION)
        self.log.info("Adding index tables.")
        for where, idx_tab in self.indices.items():
            self.log.debug("Creating index table for '%s'" % where)
            h5loc = idx_tab.h5loc
            self.log.info("  -> {0}".format(h5loc))
            indices = Table({
                "index": idx_tab.data["indices"],
                "n_items": idx_tab.data["n_items"]
            },
                            h5loc=h5loc)
            self._write_table(h5loc, indices, title='Indices')
        self.log.info(
            "Creating pytables index tables. "
            "This may take a few minutes..."
        )
        for tab in self._tables.values():
            if 'frame_id' in tab.colnames:
                tab.cols.frame_id.create_index()
            if 'slice_id' in tab.colnames:
                tab.cols.slice_id.create_index()
            if 'dom_id' in tab.colnames:
                tab.cols.dom_id.create_index()
            if 'event_id' in tab.colnames:
                try:
                    tab.cols.event_id.create_index()
                except NotImplementedError:
                    log.warning(
                        "Table '{}' has an uint64 column, "
                        "not indexing...".format(tab._v_name)
                    )
            if 'group_id' in tab.colnames:
                try:
                    tab.cols.group_id.create_index()
                except NotImplementedError:
                    log.warning(
                        "Table '{}' has an uint64 column, "
                        "not indexing...".format(tab._v_name)
                    )
            tab.flush()

        if "HDF5MetaData" in self.services:
            self.log.info("Writing HDF5 meta data.")
            metadata = self.services["HDF5MetaData"]
            for name, value in metadata.items():
                self.h5file.set_node_attr("/", name, value)

        if not self.keep_open:
            self.h5file.close()
        self.print("HDF5 file written to: {}".format(self.filename))


class HDF5Pump(Pump):
    """Read KM3NeT-formatted HDF5 files, event-by-event.

    Parameters
    ----------
    filename: str
        From where to read events. Either this OR ``filenames`` needs to be
        defined.
    filenames: list_like(str)
        Multiple filenames. Either this OR ``filename`` needs to be defined.
    skip_version_check: bool [default: False]
        Don't check the H5 version. Might lead to unintended consequences.
    ignore_hits: bool [default: False]
        If True, do not read any hit information.
    cut_mask: str
        H5 Node path to a boolean cut mask. If specified, use the boolean array
        found at this node as a mask. ``False`` means "skip this event".
        Example: ``cut_mask="/pid/survives_precut"``
    shuffle: bool, optional [default: False]
        Shuffle the group_ids, so that the blobs are mixed up.
    shuffle_function: function, optional [default: np.random.shuffle
        The function to be used to shuffle the group IDs.
    """

    def configure(self):
        self.filename = self.get('filename')
        self.filenames = self.get('filenames', default=[])
        self.skip_version_check = self.get('skip_version_check', default=False)
        self.verbose = bool(self.get('verbose'))
        self.ignore_hits = bool(self.get('ignore_hits'))
        self.cut_mask_node = self.get('cut_mask')
        self.shuffle = self.get('shuffle', default=False)
        self.shuffle_function = self.get(
            'shuffle_function', default=np.random.shuffle
        )

        self.h5file = None
        self.cut_mask = None
        self.indices = {}
        self._singletons = {}
        self.header = None
        self.group_ids = None
        self._n_groups = None
        self.index = 0

        if not self.filename and not self.filenames:
            raise ValueError("No filename(s) defined")

        if self.filename:
            self.filenames.append(self.filename)

        self.filequeue = list(self.filenames)

        self._load_next_file()

    def _load_next_file(self):
        self._reset_state()
        self._open_next_file()
        if not self.skip_version_check:
            check_version(self.h5file)
        self._read_group_info()
        if self.cut_mask_node is not None:
            self._read_cut_mask()

    def _reset_state(self):
        self._close_h5file()
        self.h5file = None
        self.cut_mask = None
        self.indices = {}
        self._singletons = {}
        self.header = None
        self.group_ids = None
        self._n_groups = None
        self.index = 0

    def _read_cut_mask(self):
        if not self.cut_mask_node.startswith('/'):
            self.cut_mask_node = '/' + self.cut_mask_node
        self.cut_mask = self.h5file.get_node(self.cut_mask_node)[:]
        mask = self.cut_mask
        if not mask.shape[0] == self.group_ids.shape[0]:
            raise ValueError("Cut mask length differs from event ids!")

    def _read_group_info(self):
        h5file = self.h5file
        if '/event_info' not in h5file and '/group_info' not in h5file:
            self.log.critical(
                "Missing /event_info or /group_info "
                "in '%s', aborting..." % h5file.filename
            )
            raise SystemExit
        elif '/group_info' in h5file:
            self.log.info("Reading group information from '/group_info'.")
            group_info = h5file.get_node('/', 'group_info')
            self.group_ids = group_info.cols.group_id[:]
            self._n_groups = len(self.group_ids)
        elif '/event_info' in h5file:
            self.log.deprecation(
                "Reading group information from '/event_info'."
            )
            event_info = h5file.get_node('/', 'event_info')
            try:
                self.group_ids = event_info.cols.group_id[:]
            except AttributeError:
                self.group_ids = event_info.cols.event_id[:]
            self._n_groups = len(self.group_ids)
        if '/raw_header' in h5file:
            try:
                self.header = HDF5Header.from_pytable(
                    h5file.get_node('/raw_header')
                )
            except TypeError:
                self.log.error("Could not parse the raw header, skipping!")
        if self.shuffle:
            self.shuffle_function(self.group_ids)

    def process(self, blob):
        try:
            blob = self.get_blob(self.index)
        except KeyError:
            self._reset_iteration()
            raise StopIteration
        self.index += 1
        return blob

    def _need_next(self, index):
        if index >= self._n_groups:
            return True

    def _open_next_file(self):
        if not self.filequeue:
            raise StopIteration("No more files available")
        if self.h5file:
            self.h5file.close()
        self.filename = self.filequeue.pop(0)
        self.h5file = tb.open_file(self.filename, 'r')
        self.print("Opening {0}".format(self.filename))
        if self.verbose:
            ("Reading %s..." % self.filename)
        return True

    def get_blob(self, index):
        if self.index >= self._n_groups:
            self._reset_iteration()
            if self.filenames:
                self._load_next_file()
            else:
                raise KeyError
        blob = Blob()
        group_id = self.group_ids[self.index]
        if self.cut_mask is not None:
            self.log.debug('Cut masks found, applying...')
            mask = self.cut_mask
            if not mask[self.index]:
                self.log.info('Cut mask blacklists this event, skipping...')
                return

        # skip groups with separate columns
        # and deal with them later
        # this should be solved using hdf5 attributes in near future
        split_table_locs = []
        ndarray_locs = []
        for tab in self.h5file.walk_nodes(classname="Table"):
            h5loc = tab._v_pathname
            loc, tabname = os.path.split(h5loc)
            if loc in split_table_locs:
                self.log.info("get_blob: '%s' is noted, skip..." % h5loc)
                continue
            if tabname == "_indices":
                self.log.debug("get_blob: found index table '%s'" % h5loc)
                split_table_locs.append(loc)
                self.indices[loc] = self.h5file.get_node(h5loc)
                continue
            if tabname.endswith("_indices"):
                self.log.debug(
                    "get_blob: found index table '%s' for NDArray" % h5loc
                )
                ndarr_loc = h5loc.replace("_indices", '')
                ndarray_locs.append(ndarr_loc)
                self.indices[ndarr_loc] = self.h5file.get_node(h5loc)
                continue
            tabname = camelise(tabname)

            index_column = None
            if 'group_id' in tab.dtype.names:
                index_column = 'group_id'
            elif 'event_id' in tab.dtype.names:
                index_column = 'event_id'

            if index_column is not None:
                try:
                    arr = tab.read_where('%s == %d' % (index_column, group_id))
                except NotImplementedError:
                    # 64-bit unsigned integer columns like ``group_id``
                    # are not yet supported in conditions
                    self.log.debug(
                        "get_blob: found uint64 column at '{}'...".
                        format(h5loc)
                    )
                    arr = tab.read()
                    arr = arr[arr[index_column] == group_id]
                except ValueError:
                    # "there are no columns taking part
                    # in condition ``group_id == 0``"
                    self.log.info(
                        "get_blob: no `%s` column found in '%s'! "
                        "skipping... " % (index_column, h5loc)
                    )
                    continue
            else:
                if h5loc not in self._singletons:
                    log.info(
                        "Caching H5 singleton: {} ({})".format(tabname, h5loc)
                    )
                    self._singletons[h5loc] = Table(
                        tab.read(),
                        h5loc=h5loc,
                        split_h5=False,
                        name=tabname,
                        h5singleton=True
                    )
                blob[tabname] = self._singletons[h5loc]
                continue

            self.log.debug("h5loc: '{}'".format(h5loc))
            blob[tabname] = Table(
                arr, h5loc=h5loc, split_h5=False, name=tabname
            )

        # skipped locs are now column wise datasets (usually hits)
        # currently hardcoded, in future using hdf5 attributes
        # to get the right constructor
        for loc in split_table_locs:
            # if some events are missing (group_id not continuous),
            # this does not work as intended
            # idx, n_items = self.indices[loc][group_id]
            idx = self.indices[loc].col('index')[self.index]
            n_items = self.indices[loc].col('n_items')[self.index]
            end = idx + n_items
            node = self.h5file.get_node(loc)
            columns = (c for c in node._v_children if c != '_indices')
            data = {}
            for col in columns:
                data[col] = self.h5file.get_node(loc + '/' + col)[idx:end]
            tabname = camelise(loc.split('/')[-1])
            blob[tabname] = Table(data, h5loc=loc, split_h5=True, name=tabname)

        if self.header is not None:
            blob['Header'] = self.header

        for ndarr_loc in ndarray_locs:
            self.log.warning("Reading %s" % ndarr_loc)
            idx = self.indices[ndarr_loc].col('index')[self.index]
            n_items = self.indices[ndarr_loc].col('n_items')[self.index]
            end = idx + n_items
            ndarr = self.h5file.get_node(ndarr_loc)
            ndarr_name = camelise(ndarr_loc.split('/')[-1])
            blob[ndarr_name] = NDArray(
                ndarr[idx:end], h5loc=ndarr_loc, title=ndarr.title
            )

        return blob

    def _reset_iteration(self):
        """Reset index to default value"""
        self.index = 0

    def __len__(self):
        return self._n_groups

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= self._n_groups:
            self._reset_iteration()
            raise StopIteration
        blob = self.get_blob(self.index)
        self.index += 1
        return blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)

        self.filename = None

    def _close_h5file(self):
        if self.h5file:
            self.h5file.close()

    def finish(self):
        self._close_h5file()


class HDF5MetaData(Module):
    """Metadata to attach to the HDF5 file.

    Parameters
    ----------
    data: dict

    """

    def configure(self):
        self.data = self.require("data")
        self.expose(self.data, "HDF5MetaData")


def convert_header_dict_to_table(header_dict):
    """Converts a header dictionary (usually from aanet) to a Table"""
    if not header_dict:
        log.warning("Can't convert empty header dict to table, skipping...")
        return
    tab_dict = defaultdict(list)
    log.debug("Param:   field_names    field_values    dtype")
    for parameter, data in header_dict.items():
        fields = []
        values = []
        types = []
        for field_name, field_value in data.items():
            fields.append(field_name)
            values.append(str(field_value))
            try:
                _ = float(field_value)    # noqa
                types.append('f4')
            except ValueError:
                types.append('a{}'.format(len(field_value)))
        tab_dict['parameter'].append(parameter)
        tab_dict['field_names'].append(' '.join(fields))
        tab_dict['field_values'].append(' '.join(values))
        tab_dict['dtype'].append(' '.join(types))
        log.debug(
            "{}: {} {} {}".format(
                tab_dict['parameter'][-1],
                tab_dict['field_names'][-1],
                tab_dict['field_values'][-1],
                tab_dict['dtype'][-1],
            )
        )
    return Table(
        tab_dict, h5loc='/raw_header', name='RawHeader', h5singleton=True
    )



# Filename: km3h5concat.py
"""
Concatenate KM3HDF5 files via pipeline.

Usage:
    km3h5concat [options] OUTFILE FILE...
    km3h5concat (-h | --help)
    km3h5concat --version

Options:
    -h --help                       Show this screen.
    --verbose                       Print more output.
    --debug                         Print everything.
    -n=NEVENTS                      Number of events; if not given, use all.
"""

from km3modules.common import StatusBar
from km3pipe import version
from km3pipe import logger

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"

log = logger.get_logger('km3pipe.io')


def km3h5concat(input_files, output_file, n_events=None, **kwargs):
    """Concatenate KM3HDF5 files via pipeline."""
    from km3pipe import Pipeline    # noqa
    from km3pipe.io import HDF5Pump, HDF5Sink    # noqa

    pipe = Pipeline()
    pipe.attach(HDF5Pump, filenames=input_files, **kwargs)
    pipe.attach(StatusBar, every=250)
    pipe.attach(HDF5Sink, filename=output_file, **kwargs)
    pipe.drain(n_events)


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    try:
        n = int(args['-n'])
    except TypeError:
        n = None

    # FILE... (ellipsis) always returns a list in docopt.
    # so the bug checking-string-length should not happen here
    infiles = args['FILE']
    outfile = args['OUTFILE']

    is_verbose = bool(args['--verbose'])
    if is_verbose:
        log.setLevel('INFO')
    is_debug = bool(args['--debug'])
    if is_debug:
        log.setLevel('DEBUG')
    km3h5concat(infiles, outfile, n)



# Filename: test_pandas.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from km3pipe.testing import TestCase

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestPandas(TestCase):
    def test_init(self):
        pass



# Filename: logger.py
# pylint: disable=locally-disabled,C0103
"""
The logging facility.

"""
from __future__ import absolute_import, print_function, division

from hashlib import sha256
import socket
import logging

from .tools import colored, supports_color

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

loggers = {}    # this holds all the registered loggers
# logging.basicConfig()

DEPRECATION = 45
logging.addLevelName(DEPRECATION, "DEPRECATION")


def deprecation(self, message, *args, **kws):
    self._log(DEPRECATION, message, args, **kws)


logging.Logger.deprecation = deprecation

if supports_color():
    logging.addLevelName(
        logging.INFO,
        "\033[1;32m%s\033[1;0m" % logging.getLevelName(logging.INFO)
    )
    logging.addLevelName(
        logging.DEBUG,
        "\033[1;34m%s\033[1;0m" % logging.getLevelName(logging.DEBUG)
    )
    logging.addLevelName(
        logging.WARNING,
        "\033[1;33m%s\033[1;0m" % logging.getLevelName(logging.WARNING)
    )
    logging.addLevelName(
        logging.ERROR,
        "\033[1;31m%s\033[1;0m" % logging.getLevelName(logging.ERROR)
    )
    logging.addLevelName(
        logging.CRITICAL,
        "\033[1;101m%s\033[1;0m" % logging.getLevelName(logging.CRITICAL)
    )
    logging.addLevelName(DEPRECATION, "\033[1;35m%s\033[1;0m" % 'DEPRECATION')


class LogIO(object):
    """Read/write logging information.
    """

    def __init__(
            self, node, stream, url='pi2089.physik.uni-erlangen.de', port=28777
    ):
        self.node = node
        self.stream = stream
        self.url = url
        self.port = port
        self.sock = None
        self.connect()

    def send(self, message, level='info'):
        message_string = "+log|{0}|{1}|{2}|{3}\r\n" \
                         .format(self.stream, self.node, level, message)
        try:
            self.sock.send(message_string)
        except socket.error:
            print("Lost connection, reconnecting...")
            self.connect()
            self.sock.send(message_string)

    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.url, self.port))


def get_logger(name):
    """Helper function to get a logger"""
    if name in loggers:
        return loggers[name]
    logger = logging.getLogger(name)
    logger.propagate = False
    pre1, suf1 = hash_coloured_escapes(name) if supports_color() else ('', '')
    pre2, suf2 = hash_coloured_escapes(name + 'salt')  \
                 if supports_color() else ('', '')
    formatter = logging.Formatter(
        '%(levelname)s {}+{}+{} '
        '%(name)s: %(message)s'.format(pre1, pre2, suf1)
    )
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    loggers[name] = logger

    return logger


def available_loggers():
    """Return a list of avialable logger names"""
    return list(logging.Logger.manager.loggerDict.keys())


def set_level(name, level):
    """Set the log level for given logger"""
    get_logger(name).setLevel(level)


def get_printer(name, color=None, ansi_code=None, force_color=False):
    """Return a function which prints a message with a coloured name prefix"""

    if force_color or supports_color():
        if color is None and ansi_code is None:
            cpre_1, csuf_1 = hash_coloured_escapes(name)
            cpre_2, csuf_2 = hash_coloured_escapes(name + 'salt')
            name = cpre_1 + '+' + cpre_2 + '+' + csuf_1 + ' ' + name
        else:
            name = colored(name, color=color, ansi_code=ansi_code)

    prefix = name + ': '

    def printer(text):
        print(prefix + str(text))

    return printer


def hash_coloured(text):
    """Return a ANSI coloured text based on its hash"""
    ansi_code = int(sha256(text.encode('utf-8')).hexdigest(), 16) % 230
    return colored(text, ansi_code=ansi_code)


def hash_coloured_escapes(text):
    """Return the ANSI hash colour prefix and suffix for a given text"""
    ansi_code = int(sha256(text.encode('utf-8')).hexdigest(), 16) % 230
    prefix, suffix = colored('SPLIT', ansi_code=ansi_code).split('SPLIT')
    return prefix, suffix



# Filename: __init__.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of commonly used modules.

"""
from __future__ import absolute_import, print_function, division

from .common import Dump, Keep, Delete, StatusBar
from .mc import RandomState



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Serial Killings -- Powerlaw or not?
===================================

Try to reproduce the analysis in `Cosma Shalizi's
blogpost <http://bactra.org/weblog/857.html>`__ with scipy.

First, import things

"""
from __future__ import absolute_import, print_function, division

import numpy as np
import pandas as pd
from scipy import stats
from statsmodels.distributions.empirical_distribution import ECDF
import matplotlib.pyplot as plt

from km3pipe.stats import bootstrap_fit
import km3pipe.style.moritz    # noqa

######################################################################
# The data (times of killings) are in a text file, let's write a small
# function to read them. We are interested in the time differences between
# the murders.
#


def get_time_deltas(fname):
    with open(fname) as f:
        raw_dates = f.read().split('\n')
    murder_dates = pd.Series(pd.to_datetime(raw_dates), name='Date')
    day = np.timedelta64(1, 'D')
    diffs = np.array(murder_dates.diff().iloc[1:-1]) / day
    diffs = pd.Series(diffs)
    return diffs


DATES_FILE = '../data/murder_dates.txt'
diffs = get_time_deltas(DATES_FILE)

######################################################################

diffs.hist(bins='auto')

######################################################################

ecdf = ECDF(diffs)
days, eucdf = ecdf.x, 1 - ecdf.y
plt.loglog(days, eucdf)
plt.xlabel('Days between murders')
plt.ylabel('Cumulative probability')

######################################################################
# Let's fit a powerlaw
#

pareto_idx, pareto_loc, pareto_scale = stats.pareto.fit(diffs)
pareto = stats.pareto(pareto_idx, pareto_loc, pareto_scale)

_ = bootstrap_fit(stats.pareto, diffs, n_iter=100)

######################################################################
# And a lognormal, because Gauss is not mocked.
#

lognorm_sig, lognorm_shape, lognorm_scale = stats.lognorm.fit(diffs)
lognorm = stats.lognorm(lognorm_sig, lognorm_shape, lognorm_scale)

_ = bootstrap_fit(stats.lognorm, diffs, n_iter=100)

######################################################################

plt.loglog(
    days,
    1 - pareto.cdf(days),
    label='Pareto Fit (exponent {:.3})'.format(pareto_idx + 1)
)
plt.loglog(days, 1 - lognorm.cdf(days), label='LogNorm Fit')
plt.loglog(days, eucdf, label='Empirical CDF')
plt.xlabel('Days between murders')
plt.ylabel('Cumulative probability')
plt.legend()



# Filename: test_testing.py
# pylint: disable=locally-disabled,C0111,R0904,C0103
from km3pipe.testing import TestCase, surrogate

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def imports():
    import my
    import my.module
    import my.module.one
    import my.module.two
    from my import module
    from my.module import one, two
    return True


class TestSurrogateModuleStubs(TestCase):
    def test_surrogating(self):
        @surrogate('my')
        @surrogate('my.module.one')
        @surrogate('my.module.two')
        def stubbed():
            imports()

        try:
            stubbed()
        except Exception as e:
            raise Exception('Modules are not stubbed correctly: %r' % e)

        with self.assertRaises(ImportError):
            imports()

    def test_context_manager(self):
        with surrogate('my'):
            with surrogate('my.module.one'):
                with surrogate('my.module.two'):
                    imports()



# Filename: test_calib.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102
from os.path import dirname, join

from km3pipe.core import Module, Pipeline
from km3pipe.dataclasses import Table
from km3pipe.hardware import Detector
from km3pipe.testing import TestCase, MagicMock, patch, skip
from km3pipe.calib import Calibration, CalibrationService

from .test_hardware import EXAMPLE_DETX

import numpy as np

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

DATA_DIR = join(dirname(__file__), '../kp-data/test_data/')
DETX_FILENAME = join(DATA_DIR, 'detx_v1.detx')


class TestCalibration(TestCase):
    """Tests for the Calibration class"""

    def test_init_with_wrong_file_extension(self):
        with self.assertRaises(NotImplementedError):
            Calibration(filename='foo')

    @patch('km3pipe.calib.Detector')
    def test_init_with_filename(self, mock_detector):
        Calibration(filename='foo.detx')
        mock_detector.assert_called_with(filename='foo.detx')

    @patch('km3pipe.calib.Detector')
    def test_init_with_det_id(self, mock_detector):
        Calibration(det_id=1)
        mock_detector.assert_called_with(
            t0set=None, calibration=None, det_id=1
        )
        Calibration(det_id=1, calibration=2, t0set=3)
        mock_detector.assert_called_with(t0set=3, calibration=2, det_id=1)

    def test_init_with_detector(self):
        det = Detector(DETX_FILENAME)
        Calibration(detector=det)

    def test_apply_to_hits_with_pmt_id(self):
        calib = Calibration(filename=DETX_FILENAME)

        hits = Table({'pmt_id': [1, 2, 1], 'time': [10.1, 11.2, 12.3]})

        chits = calib.apply(hits)

        assert len(hits) == len(chits)

        a_hit = chits[0]
        self.assertAlmostEqual(1.1, a_hit.pos_x)
        self.assertAlmostEqual(10, a_hit.t0)
        t0 = a_hit.t0
        self.assertAlmostEqual(10.1 + t0, a_hit.time)

        a_hit = chits[1]
        self.assertAlmostEqual(1.4, a_hit.pos_x)
        self.assertAlmostEqual(20, a_hit.t0)
        t0 = a_hit.t0
        self.assertAlmostEqual(11.2 + t0, a_hit.time)

    def test_apply_to_hits_with_dom_id_and_channel_id(self):
        calib = Calibration(filename=DETX_FILENAME)

        hits = Table({
            'dom_id': [2, 3, 3],
            'channel_id': [0, 1, 2],
            'time': [10.1, 11.2, 12.3]
        })

        chits = calib.apply(hits)

        assert len(hits) == len(chits)

        a_hit = chits[0]
        self.assertAlmostEqual(2.1, a_hit.pos_x)
        self.assertAlmostEqual(40, a_hit.t0)
        t0 = a_hit.t0
        self.assertAlmostEqual(10.1 + t0, a_hit.time)

        a_hit = chits[1]
        self.assertAlmostEqual(3.4, a_hit.pos_x)
        self.assertAlmostEqual(80, a_hit.t0)
        t0 = a_hit.t0
        self.assertAlmostEqual(11.2 + t0, a_hit.time)

    def test_apply_to_timeslice_hits(self):
        tshits = Table.from_template({
            'channel_id': [0, 1, 2],
            'dom_id': [2, 3, 3],
            'time': [10.1, 11.2, 12.3],
            'tot': np.ones(3, dtype=float),
            'group_id': 0,
        }, 'TimesliceHits')
        calib = Calibration(filename=DETX_FILENAME)
        c_tshits = calib.apply(tshits)
        assert len(c_tshits) == len(tshits)
        assert np.allclose([40, 80, 90], c_tshits.t0)
        # TimesliceHits is using int4 for times, so it's truncated when we pass in float64
        assert np.allclose([50.0, 91.0, 102.0], c_tshits.time)


class TestCalibrationService(TestCase):
    def test_apply_to_hits_with_dom_id_and_channel_id(self):

        hits = Table({
            'dom_id': [2, 3, 3],
            'channel_id': [0, 1, 2],
            'time': [10.1, 11.2, 12.3]
        })

        tester = self

        class HitCalibrator(Module):
            def process(self, blob):
                chits = self.services['calibrate'](hits)

                assert len(hits) == len(chits)

                a_hit = chits[0]
                tester.assertAlmostEqual(2.1, a_hit.pos_x)
                tester.assertAlmostEqual(40, a_hit.t0)
                t0 = a_hit.t0
                tester.assertAlmostEqual(10.1 + t0, a_hit.time)

                a_hit = chits[1]
                tester.assertAlmostEqual(3.4, a_hit.pos_x)
                tester.assertAlmostEqual(80, a_hit.t0)
                t0 = a_hit.t0
                tester.assertAlmostEqual(11.2 + t0, a_hit.time)
                return blob

        pipe = Pipeline()
        pipe.attach(CalibrationService, filename=DETX_FILENAME)
        pipe.attach(HitCalibrator)
        pipe.drain(1)

    def test_provided_detector_data(self):
        class DetectorReader(Module):
            def process(self, blob):
                assert 'detector' in self.services
                det = self.services['detector']
                assert isinstance(det, Detector)

        pipe = Pipeline()
        pipe.attach(CalibrationService, filename=DETX_FILENAME)
        pipe.attach(DetectorReader)
        pipe.drain(1)



# Filename: rba.py
"""
RainbowAlga online display.

Usage:
    rba console FILE DETX
    rba -t TOKEN -n EVENT_ID [-u URL] FILE
    rba (-h | --help)
    rba --version

Options:
    FILE       Input file.
    -h --help  Show this screen.

"""

from cmd import Cmd

import km3pipe as kp
from km3pipe.srv import srv_event

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def rba():
    pass


class RBAPrompt(Cmd):
    def __init__(self, filename, detx):
        Cmd.__init__(self)
        self.filename = filename
        self.pump = kp.io.GenericPump(filename)
        self.cal = kp.calib.Calibration(filename=detx)
        self.prompt = "> "
        self.current_idx = 0
        self.token = None

    # Override methods in Cmd object
    def preloop(self):
        """Initialization before prompting user for commands.

        Despite the claims in the Cmd documentaion, Cmd.preloop() is not a stub
        """
        Cmd.preloop(self)    # sets up command completion
        self._hist = []    # No history yet
        self._locals = {}    # Initialize execution namespace for user
        self._globals = {}

    def default(self, line):
        """Called on an input line when the command prefix is not recognized.
           In that case we execute the line as Python code.
        """
        pass
        # try:
        #    exec(line) in self._locals, self._globals
        # except Exception as e:
        #    print(e.__class__, ":", e)

    def do_show(self, args):
        args = args.split(' ')
        if len(args) < 2:
            print("usage: show TOKEN EVENT_ID")
            return
        self.token = args[0]
        event = int(args[1])
        self.srv_event(event)

    def do_n(self, args):
        self.current_idx += 1
        self.srv_event(self.current_idx)

    def do_p(self, args):
        if self.current_idx > 0:
            self.current_idx -= 1
            self.srv_event(self.current_idx)

    def srv_event(self, event):
        print("Serving event #{0}".format(event))
        hits = self.cal.apply(self.pump[event]["Hits"].triggered_rows)
        srv_event(self.token, hits)
        self.current_idx = event

    def do_file(self, args):
        print(self.filename)

    def do_hello(self, args):
        print("Sup?")

    def do_quit(self, args):
        self.pump.finish()
        print("Bye.")
        raise SystemExit

    def do_EOF(self, args):
        """Exit on system end of file character"""
        return self.do_quit(args)


def start_console(filename, detx):
    prompt = RBAPrompt(filename, detx)
    prompt.cmdloop('Entering RainbowAlga console...')


def main():
    from docopt import docopt
    args = docopt(__doc__)

    if args["console"]:
        start_console(args["FILE"], args["DETX"])


#    rba(arguments['FILE'])



#!/usr/bin/env python
"""Robust root->h5 converter, for rootfiles generated from I3 files.

Certain keys *have* to be skipped, otherwise the converter segfaults (!!!).

Usage:
    i3root2hdf5.py [--force] INFILE
    i3root2hdf5.py -h | --help


Options:
    -h --help     Show this screen.
    --force         Overwrite target if it already exists.
"""

from docopt import docopt

import h5py
from rootpy.io import root_open
from root_numpy import tree2array


def i3root2hdf5(infile, force=False):
    h5file = infile + '.h5'
    bad_keys = ['AntMCTree', 'MasterTree']
    rf = root_open(infile, 'r')
    keys = [k.name for k in rf.keys()]
    if force:
        mode = 'w'
    else:
        mode = 'a'
    h5 = h5py.File(h5file, mode)
    for key in keys:
        if key in bad_keys:
            continue
        tree = rf[key]
        arr = tree2array(tree)
        try:
            h5.create_dataset(
                key,
                data=arr,
                compression='gzip',
                compression_opts=9,
                shuffle=True,
                fletcher32=True,
            )
        except TypeError:
            continue
        h5.flush()
    h5.close()


def main():
    args = docopt(__doc__)
    infile = args['INFILE']
    force = args['--force']
    i3root2hdf5(infile, force=force)


if __name__ == '__main__':
    main()



# Filename: testing.py
"""
Common unit testing support for km3pipe.

"""
from __future__ import absolute_import, print_function, division

import sys
from functools import wraps
from unittest import TestCase    # noqa
from mock import MagicMock    # noqa
from mock import Mock    # noqa
from mock import patch    # noqa

from numpy.testing import assert_allclose    # noqa
import pytest    # noqa

skip = pytest.mark.skip
skipif = pytest.mark.skipif

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class surrogate(object):
    """
        Add empty module stub that can be imported
        for every subpath in path.
        Those stubs can later be patched by mock's
        patch decorator.

        This class was written by Kostia Balytskyi (ikostia @github)

        Example
        -------

        @surrogate('sys.my.cool.module1')
        @surrogate('sys.my.cool.module2')
        @mock.patch('sys.my.cool.module1', mock1)
        @mock.patch('sys.my.cool.module2', mock2)
        def function():
            from sys.my import cool
            from sys.my.cool import module1
            from sys.my.cool import module2
    """

    def __init__(self, path):
        self.path = path
        self.elements = self.path.split('.')

    def __enter__(self):
        self.prepared = self.prepare()

    def __exit__(self, *args):
        if self.prepared:
            self.restore()

    def __call__(self, func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            prepared = self.prepare()
            result = func(*args, **kwargs)
            if prepared:
                self.restore()
            return result

        return _wrapper

    @property
    def nothing_to_stub(self):
        """Check if there are no modules to stub"""
        return len(self.elements) == 0

    def prepare(self):
        """Preparations before actual function call"""
        self._determine_existing_modules()
        if self.nothing_to_stub:
            return False
        self._create_module_stubs()
        self._save_base_module()
        self._add_module_stubs()
        return True

    def restore(self):
        """Post-actions to restore initial state of the system"""
        self._remove_module_stubs()
        self._restore_base_module()

    def _get_importing_path(self, elements):
        """Return importing path for a module that is last in elements list"""
        ip = '.'.join(elements)
        if self.known_path:
            ip = self.known_path + '.' + ip
        return ip

    def _create_module_stubs(self):
        """Create stubs for all not-existing modules"""
        # last module in our sequence
        # it should be loaded
        last_module = type(
            self.elements[-1], (object, ), {
                '__all__': [],
                '_importing_path': self._get_importing_path(self.elements)
            }
        )
        modules = [last_module]

        # now we create a module stub for each
        # element in a path.
        # each module stub contains `__all__`
        # list and a member that
        # points to the next module stub in
        # sequence
        for element in reversed(self.elements[:-1]):
            next_module = modules[-1]
            module = type(
                element, (object, ), {
                    next_module.__name__: next_module,
                    '__all__': [next_module.__name__]
                }
            )
            modules.append(module)
        self.modules = list(reversed(modules))
        self.modules[0].__path__ = []

    def _determine_existing_modules(self):
        """Find out which of the modules from specified path are already
        imported (e.g. present in sys.modules) those modules should not be
        replaced by stubs.
        """
        known = 0
        while known < len(self.elements) and\
                    '.'.join(self.elements[:known + 1]) in sys.modules:
            known += 1
        self.known_path = '.'.join(self.elements[:known])
        self.elements = self.elements[known:]

    def _save_base_module(self):
        """Remember state of the last of existing modules

        The last of the sequence of existing modules
        is the only one we will change. So we must
        remember it's state in order to restore it
        afterwards.
        """

        try:
            # save last of the existing modules
            self.base_module = sys.modules[self.known_path]
        except KeyError:
            self.base_module = None

        # save `__all__` attribute of the base_module
        self.base_all = []
        if hasattr(self.base_module, '__all__'):
            self.base_all = list(self.base_module.__all__)
        if self.base_module:
            # change base_module's `__all__` attribute
            # to include the first module of the sequence
            self.base_module.__all__ = self.base_all + [self.elements[0]]
            setattr(self.base_module, self.elements[0], self.modules[0])

    def _add_module_stubs(self):
        """Push created module stubs into sys.modules"""
        for i, module in enumerate(self.modules):
            module._importing_path =\
                self._get_importing_path(self.elements[:i + 1])
            sys.modules[module._importing_path] = module

    def _remove_module_stubs(self):
        """Remove fake modules from sys.modules"""
        for module in reversed(self.modules):
            if module._importing_path in sys.modules:
                del sys.modules[module._importing_path]

    def _restore_base_module(self):
        """Restore the state of the last existing module"""
        if self.base_module:
            self.base_module.__all__ = self.base_all
            if not self.base_all:
                del self.base_module.__all__
            if hasattr(self.base_module, self.elements[0]):
                delattr(self.base_module, self.elements[0])
